/* Misc.c
 *
 * Copyright 1999, Paul Kienzle (pkienzle@cs.indiana.edu)
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * * Initial revision
 */

#define VERBOSE 0
#include <stdlib.h>
#include <math.h>
#include "stickers.h"
#include "misc.h"

/* ===================================================================== */
/* Generic routines which ought to be in a library somewhere. */
#include <sys/stat.h>
int fexist(char *name)
{
  struct stat buf;

  return lstat(name, &buf) == 0;
}

/* A routine that ought to be in the gdk library: build a cursor from
 * a text layout with '1' for foreground, '0' for background and ' '
 * for transparent.
 */
GdkCursor *gdk_cursor_new_from_data(const gchar *bits[], 
				    gint width, gint height,
				    GdkColor *fg, GdkColor *bg, 
				    gint hot_x, gint hot_y)
{
  GdkBitmap *bitmap, *mask;
  GdkCursor *cursor;
  guchar *data, *ptr;
  gint i, j;

  /* Though it does not say so on the X11 manual pages, the bitmap
   * format consists of scan lines padded to byte boundaries with the
   * bits in "reverse" order within each byte (lo bits come before hi
   * bits in the bitmap).  I assume this representation is platform
   * independent.  Let me know if it doesn't work for you. */

  /* Create space for the bitmap, padding the scanlines to byte boundaries. */
  data = g_new(guchar, ((width+7)/8)*height);

  /* Build bitmap */
  ptr = data;
  for (i=0; i < height; i++) {
    for (j=0; j < width; j++) {
      *ptr = (*ptr >> 1)|(bits[i][j]=='1'?0x80:0);
      if (j%8 == 7) ptr++;
    }
    if (j%8) *ptr++ >>= 8-j%8;
  }
  bitmap = gdk_bitmap_create_from_data(NULL, data, width, height);

  /* Build mask */
  ptr = data;
  for (i=0; i < height; i++) {
    for (j=0; j < width; j++) {
      *ptr = (*ptr >> 1)|(bits[i][j]==' '?0:0x80);
      if (j%8 == 7) ptr++;
    }
    if (j%8) *ptr++ >>= 8-j%8;
  }
  mask = gdk_bitmap_create_from_data(NULL, data, width, height);

  /* Build cursor from bitmap and mask */
  cursor = gdk_cursor_new_from_pixmap(bitmap, mask, /* Image and mask */
				      fg, bg, /* colors */ 
				      hot_x, hot_y); /* Hot point */

  /* No longer need bitmap or mask */
  gdk_pixmap_unref(bitmap);
  gdk_pixmap_unref(mask);
  g_free(data);

  return cursor;
}


#if 0
WARNING! THIS CODE HAS NEVER BEEN COMPILED OR TESTED;

/* Routine to find a free colour in an image, presumably so you can assign
 * it to transparency. It takes black if it is available, or if every
 * possible colour is used. */
#include <string.h>
#include <stdlib.h>
static int compare_color(unsigned char *rgb1, unsigned char *rgb2)
{
  if (*rgb1 < *rgb2) return -1;
  if (*rgb1++ > *rgb2++) return 1;
  if (*rgb1 < *rgb2) return -1;
  if (*rgb1++ > *rgb2++) return 1;
  if (*rgb1 < *rgb2) return -1;
  if (*rgb1++ > *rgb2++) return 1;
  return 0;
}

void gdk_imlib_find_unused_color(GdkImlibImage *im, GdkImlibColor *c)
{
  gint size = im->rgb_widht*im->rgb_height;
  guchar *colors = g_malloc(3*im->rgb_width*im->rgb_height);
  gint i;
  guchar *cp;

  memcpy(colors, im->rgb_data, size*3);
  qsort(colors, size, 3, compare_color);
  cp = colors+3;
  old=colors[0]<<16 + colors[1]<<8 + colors[2];
  if (old > 0) {
    c->r = c->g = c->b = 0;
    g_free(colors);
    return;
  }
  for(i=1; i < size; i++) {
    new=cp[0]<<16 + cp[1]<<8 + cp[2];
    if (old>=new+1) old=new;
    else {
      c->r = (old+1)>>16;
      c->g = ((old+1)>>8)&0xFF;
      c->b = (old+1)&0xFF;
      break;
    }
    cp += 3;
  }
  if (i == size) c->r = c->g = c->b = 0;

  g_free(colors);
}
#endif


/* Code for rgb_to_hsv and hsv_to_rgb taken directly from gtkcolorsel.c
 *
 * GTK - The GIMP Toolkit
 * Copyright (C) 1995-1997 Peter Mattis, Spencer Kimball and Josh MacDonald
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */ 

static void pixel_hsv_shift (guchar* rgb, gdouble hue, 
			     gdouble sat, gdouble val)
{
  gint i;
  gfloat f, h, s, v, a;
  guchar w, q, t, r, g, b, min;
  
  /* Code based on rgb_to_hsv and hsv_to_rgb from gtkcolorsel.c but no longer
   * uses true hsv instead hue is in the range [0, 720]/60, (two complete 
   * cycles of the colour circle) and v is in the range [0, 255].
   *
   * GTK - The GIMP Toolkit
   * Copyright (C) 1995-1997 Peter Mattis, Spencer Kimball and Josh MacDonald
   */

  /* convert to hsv */
  r = rgb[0]; g = rgb[1]; b = rgb[2];
  min = r;
  if (g < min) min = g;
  if (b < min) min = b;

  if (r >= g && r >= b) {
    if (r == min) h = s = 0.0;
    else {
      h = 6.0 + (gfloat)(g-b)/(gfloat)(r-min);
      s = (gfloat)(r-min)/(gfloat)r;
    }
    v = (gfloat)r;
  }
  else if (g >= b) {
    h = 8.0 + (gfloat)(b-r)/(gfloat)(g-min);
    s = (gfloat)(g-min)/(gfloat)g;
    v = (gfloat)g;
  }
  else {
    h = 10.0 + (gfloat)(r-g)/(gfloat)(b-min);
    s = (gfloat)(b-min)/(gfloat)b;
    v = (gfloat)b;
  }

  /* do transformation */
  h += hue/60.0;
  s += sat; 
  if (s >= 1.0) s = 1.0; else if (s < 0.0) s = 0.0;
  v += val*256.0; 
  if (v >= 256.0) v = 256.0; else if (v < 0.0) v = 0.0;

  /* convert back to rgb */
  i = (gint) h;
  f = h - i;
  w = v * (1.0 - s) + 0.5;
  q = v * (1.0 - (s * f)) + 0.5;
  t = v * (1.0 - (s * (1.0 - f))) + 0.5;

  switch (i%6)
    {
    case 0:
      *rgb++ = v;
      *rgb++ = t;
      *rgb++ = w;
      break;
    case 1:
      *rgb++ = q;
      *rgb++ = v;
      *rgb++ = w;
      break;
    case 2:
      *rgb++ = w;
      *rgb++ = v;
      *rgb++ = t;
      break;
    case 3:
      *rgb++ = w;
      *rgb++ = q;
      *rgb++ = v;
      break;
    case 4:
      *rgb++ = t;
      *rgb++ = w;
      *rgb++ = v;
      break;
    case 5:
      *rgb++ = v;
      *rgb++ = w;
      *rgb++ = q;
      break;
    }
}

void image_hsv_shift(GdkImlibImage *im, gdouble hue, 
		     gdouble sat, gdouble val, gdouble alpha)
{
  int i,size = im->rgb_width*im->rgb_height;
  guchar *rgb = im->rgb_data, *alpha_data = im->alpha_data;
  GdkImlibColor transparency;
  gint a;

  gdk_imlib_get_image_shape(im, &transparency);

  for (i=0; i < size; i++) {
    /* Don't translate pixel if it is transparent. */
    if (rgb[0] != transparency.r || 
	rgb[1] != transparency.g ||
	rgb[2] != transparency.b)
      pixel_hsv_shift (rgb, hue, sat, val);
    rgb += 3;
    if (alpha_data) {
      /* simply add the alpha translation */
      a = *alpha_data + (gint)rint(alpha*256.0);
      if (a >= 255) *alpha_data++ = 255; 
      else if (a <= 0) *alpha_data++ = 0;
      else *alpha_data++ = a;
    }
  }

  gdk_imlib_changed_image(im);
}


/* Single step rotation/scaling/mirroring of an image.  The rotation
 * is through an arbitrary angle expressed in radians. The scaling 
 * applies to both axes equally (no matter what borders are defined). 
 * The mirroring is vertical only.  The portion of the new image 
 * outside the orignal image will be made transparent (all four corners). 
 * Note that this cannot happen in place since the new image will be of 
 * a different size from the old. Note also that rotating and unrotating
 * an image will not yield the same image. For one thing, it will be
 * bigger since I do not crop transparent borders (create another
 * function for this purpose if you need it since there is no advantage
 * to doing it in place).  For another thing, the truncation from
 * real coordinates to integer coordinates may cause some pixels to shift.
 */
GdkImlibImage *image_reshape(GdkImlibImage *im, 
		gfloat scale, gfloat angle, gboolean flip)
{
  /* Compute rotation matrix */
  const gfloat sin_th = sin(angle);
  const gfloat cos_th = cos(angle);
  const gfloat scaled_sin_th = sin_th/scale;
  const gfloat scaled_cos_th = cos_th/scale;
  const gfloat width = im->rgb_width;
  const gfloat height = im->rgb_height;

  gfloat x, y, tx, ty, x1, y1, x2, y2;
  guchar *src;
  int new_w, new_h, i, j;
  guchar *rgb_data, *alpha_data, *rgb_ptr, *alpha_ptr;
  GdkImlibColor transparency;
  guchar transparency_src[3];
  
  /* Simple optimization: if there is no rotation or scaling, simply
     clone and possibly flip the image. */
  if (fabs(angle) < 1e-6 && fabs(scale-1.0) < 1e-6) {
    im = gdk_imlib_clone_image(im);
    if (flip) gdk_imlib_flip_image_horizontal(im);
    return im;
  }

  /* Compute new image size by scaling and rotating two of the corners
   * and picking the maximum dimension. Since we are rotating about the 
   * center, all the corners are going to be at (+- a*w/2, +- a*h/2).
   * We need the absolute value since the new corner may be anywhere 
   * in the circle.  We need to multiply the result by two since the 
   * coordinates are at the radius, not the diameter. */
  x1 = width*cos_th + height*sin_th,  y1 = height*cos_th - width*sin_th;
  x2 = -width*cos_th + height*sin_th, y2 = height*cos_th + width*sin_th;
  new_w = (int)(scale*MAX(fabs(x1),fabs(x2))-1.0+1e-8);
  new_h = (int)(scale*MAX(fabs(y1),fabs(y2))-1.0+1e-8);
  TRACE("angle %f, scale %f, width %f, height %f -> (%d,%d)", 
	  angle, scale, width, height, new_w, new_h);
  if (new_w <= 0) new_w = 1;
  if (new_h <= 0) new_h = 1;

  /* Create new space for the new image */
  rgb_ptr = rgb_data = malloc(new_w*new_h*3);
  alpha_ptr = alpha_data = im->alpha_data?malloc(new_w*new_h):NULL;

  /* Create a transparent cell for the parts that weren't on the
     original image. Note that if there is no transparent color
     defined, white will be used. */
  gdk_imlib_get_image_shape(im, &transparency);
  transparency_src[0] = transparency.r & 0xFF;
  transparency_src[1] = transparency.g & 0xFF;
  transparency_src[2] = transparency.b & 0xFF;

  /* Scale/rotation by inverse transformation method.  Step through 
   * the cells of the new image one by one, computing the unrotated 
   * and unscaled coordinates to locate a pixel in the original image.
   *
   * To do this, shift the pixel coordinate origin from the top-left
   * to the center (i,j)->(i-w'/2,j-h'/2), multiply by the rotation 
   * matrix for negative theta:
   *   / cos th  sin th\ / i-w'/2 \
   *   \-sin th  cos th/ \ j-h'/2 /
   * which gives:
   *   / i*cos(th) + j*sin(th) - w'/2*cos(th) - h'/2*sin(th) \
   *   \ j*cos(th) - i*sin(th) - h'/2*cos(th) + w'/2*sin(th) /
   * Undo the scaling:
   *   / i*cos(th)/a + j*sin(th)/a - w'/2*cos(th)/a - h'/2*sin(th)/a \
   *   \ j*cos(th)/a - i*sin(th)/a - h'/2*cos(th)/a + w'/2*sin(th)/a /
   * And finally translate the pixel coordinate origin form the center 
   * to the top-left of the original image by adding (w/2,h/2):
   *   / i*cos(th)/a + j*sin(th)/a - w'/2*cos(th)/a - h'/2*sin(th)/a + w/2 \
   *   \ j*cos(th)/a - i*sin(th)/a - h'/2*cos(th)/a + w'/2*sin(th)/a + h/2 /
   *
   * The inner loop over the transformed rows is much faster if you
   * start with
   *   tx <- j*sin(th)/a - w'/2*cos(th)/a - h'/2*sin(th)/a + w/2
   *   ty <- j*cos(th)/a - h'/2*cos(th)/a - w'/2*sin(th)/a + h/2
   * and increment by (cos(th)/a, -sin(th)/a)
   *   tx <- tx + cos(th)/a
   *   ty <- ty - sin(th)/a
   * Similarly, the outer loop is faster if you initialize tx,ty with x,y
   *   x <- -w'/2*cos(th)/a - h'/2*sin(th)/a + w/2
   *   y <- -w'/2*cos(th)/a - h'/2*sin(th)/a + w/2
   * and increment by (sin(th)/a, cos(th)/a)
   *   x <- x + sin(th)/a
   *   y <- y + cos(th)/a
   * We needn't be terrible concerned with round-off since we are
   * adding similar-sized values every step of the way, since there
   * aren't that many steps, and since we are rounding to the nearest
   * integer when we are done anyway.
   *
   * We can mirror the rotated image by scanning the transformed rows
   * from back to front.  That is, start at the coordinates (i+w'/2,j-h'/2), 
   * and decrement each time.  The pixels are still being assigned from
   * 0 up to w', so the resulting image is reversed.
   * */

  /* Source for pixel 0,0 in the transformed image */
  if (flip) {
    x = 0.5 * (width + new_w * scaled_cos_th - new_h * scaled_sin_th);
    y = 0.5 * (height - new_h * scaled_cos_th - new_w * scaled_sin_th);
  }
  else {
    x = 0.5 * (width - new_w * scaled_cos_th - new_h * scaled_sin_th);
    y = 0.5 * (height - new_h * scaled_cos_th + new_w * scaled_sin_th);
  }

  /* Assign transformed pixels, row by row */
  for (j=0; j < new_h; j++) {
    /* Move to the source pixel for the start of the scan line */
    tx = x; 
    ty = y;
    for (i = 0; i < new_w; i++) {
      /* Determine if source pixel is within the source image */
      if (tx < 0.0 || tx >= width || ty < 0.0 || ty >= height) {
	src = transparency_src;
	if (alpha_data) *alpha_ptr++ = 0;
      } else {
	src = &im->rgb_data[3*(((int)ty) * im->rgb_width + (int)tx)];
	if (alpha_data)
	  *alpha_ptr++ = im->alpha_data[((int)ty)*im->rgb_width+(int)tx];
      }
      
      /* Update the image pixel */
      *rgb_ptr++ = src[0];
      *rgb_ptr++ = src[1];
      *rgb_ptr++ = src[2];

      /* Move to next source pixel for current scan line */
      if (flip) {
	tx -= scaled_cos_th;
	ty += scaled_sin_th;
      } else {
	tx += scaled_cos_th;
	ty -= scaled_sin_th;
      }
    }
    
    /* Move the start of the next scan line */
    x += scaled_sin_th;
    y += scaled_cos_th;
  }

  im = gdk_imlib_create_image_from_data(rgb_data, alpha_data, new_w, new_h);
  gdk_imlib_set_image_shape(im, &transparency);
  free(rgb_data); if (alpha_data) free(alpha_data);
  return im;
}


/* ===================================================================== */
/* Pointer confinement */
static GList *confine_list=NULL;

void confine_add(GtkWidget *widget)
{
  confine_list = g_list_prepend(confine_list, widget->window);
}

void confine_remove(GtkWidget *widget)
{
  confine_list = g_list_remove(confine_list, widget->window);
}

void confine_pointer(GtkWidget *widget)
{
  GtkWidget *root;
  GdkWindow *pointer_window;
  gint err;

  /* Locate root of widget tree, and the location of the pointer */
  root = gtk_widget_get_toplevel(widget);
  pointer_window = gdk_window_get_pointer(root->window,NULL,NULL,NULL);
  if (pointer_window == NULL) pointer_window = root->window;

  /* Release any current grabs before grabbing again */
  if (gdk_pointer_is_grabbed()) gdk_pointer_ungrab(GDK_CURRENT_TIME);

  /* If the pointer is in the canvas, we want to avoid a redraw, so let
   * the canvas keep it. It will give back the pointer when it leaves. */
  if (g_list_find(confine_list, pointer_window) && widget == root) {
    err = gdk_pointer_grab (pointer_window, TRUE, 0,
			    widget->window, NULL, GDK_CURRENT_TIME);
    if (err) TRACE("canvas confine_cursor: pointer grab=%d", err); 
  }
  else {
    err = gdk_pointer_grab (widget->window, TRUE, 0, 
			    widget->window, NULL, GDK_CURRENT_TIME);
    if (err) TRACE("widget confine_cursor: pointer grab=%d", err); 
    
    /* Force entry into the window containing the pointer */
    if (pointer_window!=widget->window && widget == root) {
      gdk_window_hide(pointer_window);
      gdk_window_show(pointer_window);
    }
  }
}


#ifndef HAVE_GRAB_IMAGE
/*
-------------------------------------------------------------------------------
                                I   M   L   I   B
                                      1.9.4
-------------------------------------------------------------------------------

This software is Copyright (C) 1998 By The Rasterman (Carsten Haitzler). I
accept no responsability for anythign this software may or may not do to
your system - you use it completely at your own risk. This software comes
under the LGPL and GPL licences. The library itself is LGPL (all software in
Imlib and gdk_imlib directories) see the COPYING and COPYING.LIB files for
full legal details.

-------------------------------------------------------------------------------
*/
#include <gdk/gdkx.h>
#include <gdk_imlib_private.h>
#ifdef HAVE_SHM
#include <sys/shm.h>
#include <X11/extensions/XShm.h>
#endif
#if 0
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Xatom.h>
#include <X11/Xos.h>
#include <X11/extensions/shape.h>
#include <X11/cursorfont.h>
#endif

static char         x_error = 0;
static void
__handle_x_error(Display * d, XErrorEvent * ev)
{
  d = NULL;
  ev = NULL;
  x_error = 1;
}

GdkImlibImage      *
gdk_imlib_create_image_from_drawable(GdkWindow * gwin, GdkBitmap * gmask,
				     int x, int y, int width, int height)
{
  unsigned char      *data = NULL, *ptr, r, g, b;
  unsigned long       pixel;
  int                 i, xx, yy, w, h, inx, iny, clipx, clipy, rx, ry;
  XImage             *xim;

#ifdef HAVE_SHM
  XShmSegmentInfo     shminfo;

#endif
  XWindowAttributes   xatt, ratt;
  Colormap            cmap;
  static char         shm_checked = 0, shm = 1;
  XErrorHandler       erh = NULL;
  Window              chld;
  char                is_pixmap = 0;
  Visual             *vis;
  GdkImlibImage      *im;
  GdkImlibColor       ctab[256];
  Display            *disp;
  Window              win;
  Pixmap              mask;

  inx = 0;
  iny = 0;
  w = width;
  h = height;
  vis = id->x.visual;
  disp = id->x.disp;
  win = GDK_WINDOW_XWINDOW(gwin);
  if (gmask)
    mask = GDK_WINDOW_XWINDOW(gmask);
  else
    mask = 0;
  XGrabServer(disp);
  erh = XSetErrorHandler((XErrorHandler) __handle_x_error);
  x_error = 0;
  XGetWindowAttributes(disp, win, &xatt);
  XFlush(disp);
  if (x_error)
    {
      x_error = 0;
      is_pixmap = 1;
      XGetGeometry(disp, win, &chld, &rx, &rx,
		   (unsigned int *)&xatt.width, (unsigned int *)&xatt.height,
		   (unsigned int *)&rx, (unsigned int *)&xatt.depth);
      XFlush(disp);
      if (x_error)
	{
	  XUngrabServer(disp);
	  XFlush(disp);
	  XSetErrorHandler((XErrorHandler) erh);
	  return NULL;
	}
    }
  XSetErrorHandler((XErrorHandler) erh);
  if (!is_pixmap)
    {
      XGetWindowAttributes(disp, xatt.root, &ratt);
      XTranslateCoordinates(disp, win, xatt.root, 0, 0, &rx, &ry, &chld);
      if ((xatt.map_state != IsViewable) &&
	  (xatt.backing_store == NotUseful))
	{
	  XUngrabServer(disp);
	  XFlush(disp);
	  return NULL;
	}
    }
  clipx = 0;
  clipy = 0;

  x = x - inx;
  y = y - iny;

  width = xatt.width - x;
  height = xatt.height - y;
  if (width > w)
    width = w;
  if (height > h)
    height = h;

  if (!is_pixmap)
    {
      if ((rx + x + width) > ratt.width)
	width = ratt.width - (rx + x);
      if ((ry + y + height) > ratt.height)
	height = ratt.height - (ry + y);
    }
  if (x < 0)
    {
      clipx = -x;
      width += x;
      x = 0;
    }
  if (y < 0)
    {
      clipy = -y;
      height += y;
      y = 0;
    }
  if (!is_pixmap)
    {
      if ((rx + x) < 0)
	{
	  clipx -= (rx + x);
	  width += (rx + x);
	  x = -rx;
	}
      if ((ry + y) < 0)
	{
	  clipy -= (ry + y);
	  height += (ry + y);
	  y = -ry;
	}
    }
  if ((width <= 0) || (height <= 0))
    {
      XUngrabServer(disp);
      XSync(disp, False);
      return NULL;
    }
#ifdef HAVE_SHM
  if (shm)
    {
      if (!shm_checked)
	{
	  erh = XSetErrorHandler((XErrorHandler) __handle_x_error);
	}
      xim = XShmCreateImage(disp, vis, xatt.depth, ZPixmap, NULL,
			    &shminfo, width, height);
      if (!shm_checked)
	{
	  XSync(disp, False);
	  if (x_error)
	    {
	      shm = 0;
	      XDestroyImage(xim);
	      xim = XGetImage(disp, win, x, y, width, height, 0xffffffff, ZPixmap);
	      XSetErrorHandler((XErrorHandler) erh);
	      shm_checked = 1;
	    }
	  else
	    {
	      shminfo.shmid = shmget(IPC_PRIVATE, xim->bytes_per_line *
				     xim->height, IPC_CREAT | 0666);
	      if (shminfo.shmid < 0)
		{
		  shm = 0;
		  XDestroyImage(xim);
		  xim = XGetImage(disp, win, x, y, width, height, 0xffffffff, ZPixmap);
		  XSetErrorHandler((XErrorHandler) erh);
		  shm_checked = 1;
		}
	      else
		{
		  shminfo.shmaddr = xim->data = shmat(shminfo.shmid, 0, 0);
		  if (shminfo.shmaddr == (char *)-1)
		    {
		      shm = 0;
		      XDestroyImage(xim);
		      shmctl(shminfo.shmid, IPC_RMID, 0);
		      xim = XGetImage(disp, win, x, y, width, height, 0xffffffff, ZPixmap);
		      XSetErrorHandler((XErrorHandler) erh);
		      shm_checked = 1;
		    }
		  else
		    {
		      shminfo.readOnly = False;
		      XShmAttach(disp, &shminfo);
		    }
		}
	    }
	}
      else
	{
	  shminfo.shmid = shmget(IPC_PRIVATE, xim->bytes_per_line *
				 xim->height, IPC_CREAT | 0666);
	  if (shminfo.shmid < 0)
	    {
	      shm = 0;
	      XDestroyImage(xim);
	      xim = XGetImage(disp, win, x, y, width, height, 0xffffffff, ZPixmap);
	      XSetErrorHandler((XErrorHandler) erh);
	      shm_checked = 1;
	    }
	  else
	    {
	      shminfo.shmaddr = xim->data = shmat(shminfo.shmid, 0, 0);
	      if (shminfo.shmaddr == (char *)-1)
		{
		  shm = 0;
		  XDestroyImage(xim);
		  shmctl(shminfo.shmid, IPC_RMID, 0);
		  xim = XGetImage(disp, win, x, y, width, height, 0xffffffff, ZPixmap);
		  XSetErrorHandler((XErrorHandler) erh);
		  shm_checked = 1;
		}
	      else
		{
		  shminfo.readOnly = False;
		  XShmAttach(disp, &shminfo);
		}
	    }
	}
      if (!shm_checked)
	{
	  XSync(disp, False);
	  if (x_error)
	    {
	      shm = 0;
	      XDestroyImage(xim);
	      xim = XGetImage(disp, win, x, y, width, height, 0xffffffff, ZPixmap);
	      shm_checked = 1;
	    }
	  XSetErrorHandler((XErrorHandler) erh);
	  shm_checked = 1;
	}
    }
  else
#endif /* HAVE_SHM */
    xim = XGetImage(disp, win, x, y, width, height, 0xffffffff, ZPixmap);
#ifdef HAVE_SHM
  if (shm)
    XShmGetImage(disp, win, xim, x, y, 0xffffffff);
#endif
  XUngrabServer(disp);
  XFlush(disp);

  if (xatt.depth == 1)
    {
      ctab[0].r = 255;
      ctab[0].g = 255;
      ctab[0].b = 255;
      ctab[1].r = 0;
      ctab[1].g = 0;
      ctab[1].b = 0;
    }
  if (xatt.depth <= 8)
    {
      XColor              cols[256];

      if (!is_pixmap)
	{
	  cmap = xatt.colormap;
	  if (cmap == None)
	    cmap = id->x.root_cmap;
	}
      else
	cmap = id->x.root_cmap;

      for (i = 0; i < (1 << id->x.depth); i++)
	{
	  cols[i].pixel = i;
	  cols[i].flags = DoRed | DoGreen | DoBlue;
	}
      XQueryColors(disp, cmap, cols, 1 << id->x.depth);
      for (i = 0; i < (1 << id->x.depth); i++)
	{
	  ctab[i].r = cols[i].red >> 8;
	  ctab[i].g = cols[i].green >> 8;
	  ctab[i].b = cols[i].blue >> 8;
	  ctab[i].pixel = cols[i].pixel;
	}
    }
  data = malloc(width * height * 3);
  if (data)
    {
      ptr = data;
      switch (xatt.depth)
	{
	case 0:
	case 1:
	case 2:
	case 3:
	case 4:
	case 5:
	case 6:
	case 7:
	case 8:
	  for (yy = 0; yy < height; yy++)
	    {
	      for (xx = 0; xx < width; xx++)
		{
		  pixel = XGetPixel(xim, xx, yy);
		  r = ctab[pixel & 0xff].r;
		  g = ctab[pixel & 0xff].g;
		  b = ctab[pixel & 0xff].b;
		  *ptr++ = r;
		  *ptr++ = g;
		  *ptr++ = b;
		}
	    }
	  break;
	case 16:
	case 15:
	  if (id->x.render_depth == 16)
	    {
	      for (yy = 0; yy < height; yy++)
		{
		  for (xx = 0; xx < width; xx++)
		    {
		      pixel = XGetPixel(xim, xx, yy);
		      r = (pixel >> 8) & 0xf8;
		      g = (pixel >> 3) & 0xfc;
		      b = (pixel << 3) & 0xf8;
		      *ptr++ = r;
		      *ptr++ = g;
		      *ptr++ = b;
		    }
		}
	    }
	  else
	    {
	      for (yy = 0; yy < height; yy++)
		{
		  for (xx = 0; xx < width; xx++)
		    {
		      pixel = XGetPixel(xim, xx, yy);
		      r = (pixel >> 7) & 0xf8;
		      g = (pixel >> 3) & 0xf8;
		      b = (pixel << 3) & 0xf8;
		      *ptr++ = r;
		      *ptr++ = g;
		      *ptr++ = b;
		    }
		}
	    }
	  break;
	case 24:
	case 32:
	  for (yy = 0; yy < height; yy++)
	    {
	      for (xx = 0; xx < width; xx++)
		{
		  pixel = XGetPixel(xim, xx, yy);
		  r = (pixel >> 16) & 0xff;
		  g = (pixel >> 8) & 0xff;
		  b = pixel & 0xff;
		  *ptr++ = r;
		  *ptr++ = g;
		  *ptr++ = b;
		}
	    }
	  break;
	default:
	  for (yy = 0; yy < height; yy++)
	    {
	      for (xx = 0; xx < width; xx++)
		{
		  r = rand();
		  g = rand();
		  b = rand();
		  *ptr++ = r;
		  *ptr++ = g;
		  *ptr++ = b;
		}
	    }
	  break;
	}
    }

#ifdef HAVE_SHM
  if (shm)
    {
      XSync(disp, False);
      XShmDetach(disp, &shminfo);
      shmdt(shminfo.shmaddr);
      shmctl(shminfo.shmid, IPC_RMID, 0);
    }
#endif
  XDestroyImage(xim);

  if (data)
    {
      im = gdk_imlib_create_image_from_data(data, NULL, width, height);
      free(data);
      return im;
    }
  else
    return NULL;
}

#endif /* NEED_GRAB_IMAGE */

